/********************************************************************************
 * ServerAI.java
 * Copyright (c) 2002 by Stefan Buettcher
 *
 * This class represents a server-side Artificial Intelligence that automatically
 * offers games and plays games.
 *
 *******************************************************************************/

import java.net.*;
import java.io.*;
import java.util.*;

public class ServerAI extends Player {
  public static final int CONNECT4_AI = 1;
  public static final int CHECKERS_AI = 2;

  private String gameName;
  private Game owner;
  private Random random;

  private AI_Module ai;

  public ServerAI(String aName, int aType, Game owner) {
  // just an ordinary constructor
    name = aName;
    match = null;
    canStart = false;
    connection = null;
    this.owner = owner;
    random = new Random(System.currentTimeMillis() + 7);

    if (aType == CONNECT4_AI) {
      gameName = "4Gewinnt";
      ai = new Connect4AI(this);
      ai.start();
    }
    else if (aType == CHECKERS_AI) {
      gameName = "Checkers";
      ai = new CheckersAI(this);
      ai.start();
    }
    else {
      gameName = null;
      ai = null;
    }
  }

  public void setCanStart(boolean can) {
    canStart = can;
  }

  public boolean getCanStart() {
    return canStart;
  }

  public void setMatch(Match aMatch) {
    match = aMatch;
  }

  public Match getMatch() {
    return match;
  }

  public void setConnection(Connection conn) {
  }

  public void sendMessage(String message) {
    if ((message.indexOf("has won") > 0) || (message.indexOf("have won") > 0)) {
      if (match != null)
        sendToServer("quit " + match.getName());
      sendToServer("offer " + gameName + "_AI_" + (random.nextInt(89) + 10));
      sendToServer("canstart=true");
      return;
    }
    if ((message.startsWith("[NEWS] ")) && (ai != null)) {
      message = message.substring(7);
      if (message.startsWith("You are"))
        ai.setWho(message.charAt(message.length() - 2));
      else if (message.startsWith("board="))
        ai.setBoard(message.substring(6));
      else if (message.startsWith("Please move."))
        ai.requestMove();
      else if (message.startsWith("Chat from ")) {
        message = message.substring(10);
        ai.requestStupidMessage();
      }
    }
  }

  public void sendToServer(String message) {
    owner.processMessage(this, message);
  }

  public String getName() {
    return name;
  }

  public String toString() {
    return name;
  }

  public int compareTo(Object o) {
  // Player instances must be comparable if we want to organize them in a tree
    return name.compareTo(o.toString());
  }

  public String getGameName() {
    return gameName;
  }

}


abstract class AI_Module extends Thread {
  public abstract void setWho(char who);
  public abstract void setBoard(String boardState);
  public abstract void requestMove();
  public abstract void requestStupidMessage();
}


class Connect4AI extends AI_Module {
  private final int INTERVAL = 200;
  private final int SEARCH_DEPTH = 6;
  private ServerAI owner;
  public boolean terminated, moveRequested, started, chatRequested;
  private Random random;
  private char who;
  private String board;
  private char b[][];

  public Connect4AI(ServerAI owner) {
    this.owner = owner;
    terminated = false;
    random = new Random(System.currentTimeMillis() + 11);
    who = '_';
    board = null;
    moveRequested = false;
    chatRequested = false;
    started = false;
  }

  public void setWho(char who) {
    this.who = who;
  }

  public void setBoard(String boardState) {
    board = boardState;
  }

  public void requestMove() {
    moveRequested = true;
  }

  public void requestStupidMessage() {
    chatRequested = true;
  }

  public void run() {
    while (!terminated) {
      if (!started) {
        started = true;
        try { Thread.sleep(1000); }
          catch (Exception ex) { }
        owner.sendToServer("offer AI_Game" + (random.nextInt(89) + 10));
        owner.sendToServer("canstart=true");
      }
      try {
        Thread.sleep(INTERVAL);
        if (moveRequested) {
          moveRequested = false;
          String b = board;
          board = null;
          getMove(b);
        }
        else if (chatRequested) {
	  chatRequested = false;
          switch (random.nextInt(10)) {
            case 0: owner.sendToServer("chat Nice to talk to you."); break;
            case 1: owner.sendToServer("chat Let's talk about the weather."); break;
            case 2: owner.sendToServer("chat Haven't you seen that nice huttlepaffody over there?"); break;
            case 3: owner.sendToServer("chat I believe the most beatiful part of a woman is the boobies."); break;
            case 4: owner.sendToServer("chat Oh, my goodness!"); break;
            case 5: owner.sendToServer("chat You should rather concentrate on this game than sending stupid messages around the world."); break;
            case 6: owner.sendToServer("chat Do we know each other?"); break;
	    case 7: owner.sendToServer("Ich bin ein krasser 4-Gewinnt-Meister."); break;
	    case 8: owner.sendToServer("Meinst du, du hast eine Chance gegen einen Checker wie mich?"); break;
	    case 9: owner.sendToServer("Blupp."); break;
          }
	}
      }
      catch (Exception e) {
      }
    }
  }

  private void getMove(String boardState) {
    int x, y;
    b = new char[7][6];
    if ((boardState != null) && (boardState.length() == 42)) {
      for (x = 0; x < 7; x++)
        for (y = 0; y < 6; y++)
          b[x][y] = boardState.charAt(x + y * 7);
      int move = getMoveInternal();
      if (move >= 0)
        owner.sendToServer("move " + move);
    }
  }

  private int neuralValue(char who) {
    return 0;
  }

  private int boardValue(char who, int parentMax, int depth) {
    if (isEndState())
      return +100;
    if (depth >= SEARCH_DEPTH)
      return neuralValue(who);
    else {
      int value;
      int max = -200;
      if (who == 'w')
        who = 'b';
      else
        who = 'w';
      Vector moves = getPossibleMoves();
      for (int cnt = 0; cnt < moves.size(); cnt++) {
        int move = ((Integer)moves.elementAt(cnt)).intValue();
        doMove(move, who);
        value = boardValue(who, max, depth + 1);
        if (value > max) {
          max = value;
          if (max >= -parentMax) {
            undoMove(move);
            max += 1;
            break;
          }
        }
        undoMove(move);
      }
      if (who == 'w')
        who = 'b';
      else
        who = 'w';

      return -max;
    }
  }

  private boolean row(int x, int y, int count, int xInc, int yInc, char piece) {
  // returns true if there is a row in the direction indicated by (xInc, yInc)
  // of length >= count of pieces of type piece
    if (count <= 0)
      return true;
    if ((x < 0) || (x >= 7) || (y < 0) || (y >= 6))
      return false;
    return (b[x][y] == piece) &
           (row(x + xInc, y + yInc, count - 1, xInc, yInc, piece));
  }

  public boolean isEndState() {
  // returns true if the current board state is a final state
    int x, y, winner;
    for (y = 0; y < 6; y++)
      for (x = 0; x < 7; x++) {
        char c = b[x][y];
        if (c != '.') {
          if (row(x, y, 4, 1, 0, c))
            return true;
          if (row(x, y, 4, 0, 1, c))
            return true;
          if (row(x, y, 4, 1, 1, c))
            return true;
          if (row(x, y, 4, -1, 1, c))
            return true;
        }
      }
    return false;
  }

  private void doMove(int move, char who) {
    int y = 5;
    while ((y >= 0) && (b[move][y] != '.'))
      y--;
    if (y >= 0)
      b[move][y] = who;
  }

  private void undoMove(int move) {
    int y = 0;
    while ((y < 5) && (b[move][y] == '.'))
      y++;
    if (y <= 5)
      b[move][y] = '.';
  }

  private int getMoveInternal() {
    Vector moves = getPossibleMoves();
    int max = -200;
    int result = -1;
    int value, move;
    for (int i = 0; i < moves.size(); i++) {
      move = ((Integer)moves.elementAt(i)).intValue();
      doMove(move, who);
      value = boardValue(who, max, 0);
      if (value >= max) {
        if ((result < 0) || (value > max))
          result = move;
        else if (random.nextBoolean())
          result = move;
        max = value;
        if (max == 100) break;
      }
      undoMove(move);
    }
    return result;
  }

  private Vector getPossibleMoves() {
    int x;
    Vector result = new Vector();
    for (x = 0; x < 7; x++)
      if (b[x][0] == '.')
        result.add(new Integer(x));
    return result;
  }

}


class CheckersAI extends AI_Module {
  private final int INTERVAL = 200;
  private final int SEARCH_DEPTH = 6;
  private ServerAI owner;
  public boolean terminated, moveRequested, started, chatRequested;
  private Random random;
  private char who, color, enemy;
  private String boardStr;
  private char board[][];

  public CheckersAI(ServerAI owner) {
    this.owner = owner;
    terminated = false;
    random = new Random(System.currentTimeMillis() + 29);
    who = '_';
    board = null;
    moveRequested = false;
    chatRequested = false;
    started = false;
  }

  public void setWho(char who) {
    this.who = who;
  }

  public void setBoard(String boardState) {
    boardStr = boardState;
  }

  public void requestMove() {
    moveRequested = true;
  }

  public void requestStupidMessage() {
    chatRequested = true;
  }

  public void run() {
    while (!terminated) {
      if (!started) {
        started = true;
        try { Thread.sleep(1000); }
          catch (Exception ex) { }
        owner.sendToServer("offer AI_Game" + (random.nextInt(89) + 10));
        owner.sendToServer("canstart=true");
      }
      try {
        Thread.sleep(INTERVAL);
        if (moveRequested) {
          moveRequested = false;
          String b = boardStr;
          boardStr = null;
          getMove(b);
        }
        else if (chatRequested) {
	  chatRequested = false;
          switch (random.nextInt(10)) {
            case 0: owner.sendToServer("chat Hi there, checker!."); break;
            case 1: owner.sendToServer("chat Let's talk about the weather."); break;
            case 2: owner.sendToServer("chat Haven't you seen that nice huttlepaffody over there?"); break;
            case 3: owner.sendToServer("chat I believe the most beatiful part of a woman is the boobies."); break;
            case 4: owner.sendToServer("chat Oh, my goodness!"); break;
            case 5: owner.sendToServer("chat You should rather concentrate on this game than sending stupid messages around the world."); break;
            case 6: owner.sendToServer("chat Do we know each other?"); break;
            case 7: owner.sendToServer("Ich bin ein krasser 4-Gewinnt-Meister."); break;
            case 8: owner.sendToServer("Meinst du, du hast eine Chance gegen einen Checker wie mich?"); break;
            case 9: owner.sendToServer("Blupp."); break;
          }
        }
      }
      catch (Exception e) {
      }
    }
  }

  private void getMove(String boardState) {
    color = who;
    if (color == 'w') enemy = 'b';
      else enemy = 'w';
    setBoardInternal(boardState);
    Move m = getMoveInternal(board);
    if (m != null)
      owner.sendToServer("move " + m.toString());
  }

  private char[][] duplicateBoard(char board[][]) {
    char result[][] = new char[8][8];
    for (int x = 0; x < 8; x++)
      for (int y = 0; y < 8; y++)
        result[x][y] = board[x][y];
    return result;
  }

  private int neuralValue(char board[][]) {
    int x, y, value;
    char c;
    int result = 0;
    for (x = 0; x < 8; x++)
      for (y = 0; y < 8; y++) {
        c = board[x][y];
        if (c != '.') {
          value = 0;
          if (c == 'W') value = +200;
          if (c == 'B') value = -200;
          if (c == 'w') value = + (100 + (y + 1) * (y + 1));
          if (c == 'b') value = - (100 + (7 - y + 1) * (7 - y + 1));
          if (color == 'w')
            result += value;
          else
            result -= value;
        }
      }
    return result;
  }

  private int boardValue(char board[][], int depth, int parentMax) {
    int x, y, w, b;

    char board2[][];

    b = w = 0;
    for (x = 0; x < 8; x++)
      for (y = 0; y < 8; y++) {
        if ((board[x][y] == 'w') || (board[x][y] == 'W'))
          w++;
        if ((board[x][y] == 'b') || (board[x][y] == 'B'))
          b++;
      }
    if (w == 0)
      return +1000;
    if (b == 0)
      return +1000;

    if (depth >= SEARCH_DEPTH)
      return neuralValue(board);
    else {
      int value;
      int max = -1000;

      if (color == 'w') color = 'b';
        else color = 'w';
      if (enemy == 'w') enemy = 'b';
        else enemy = 'w';

      Vector moves = getPossibleMoves(board);
      for (int cnt = 0; cnt < moves.size(); cnt++) {
        Move m = (Move)moves.elementAt(cnt);

        board2 = duplicateBoard(board);
        doMove(board2, m);

        value = boardValue(board2, depth + 1, max);
        if (value > max) {
          max = value;
          if (max > -parentMax)
            break;
        }
      }

      if (color == 'w') color = 'b';
        else color = 'w';
      if (enemy == 'w') enemy = 'b';
        else enemy = 'w';

      return -max;
    }
  }

  private void doMove(char board[][], Move m) {
    int x, y;
    for (int i = 1; i < m.length; i++) {
      x = m.getX(i - 1);
      y = m.getY(i - 1);
      board[x][y] = '.';
      board[(x + m.getX(i)) / 2][(y + m.getY(i)) / 2] = '.';
    }
  }

  private void setBoardInternal(String boardStr) {
    int x, y;
    board = new char[8][8];
    if ((boardStr == null) || (boardStr.length() != 64))
      return;
    for (x = 0; x < 8; x++)
      for (y = 0; y < 8; y++)
        board[x][y] = boardStr.charAt(x + y * 8);
  }

  private Move getMoveInternal(char board[][]) {
    int cnt, value;
    int max = -1000;

    char board2[][];

    Move result = null;
    Vector moves = getPossibleMoves(board);
    for (cnt = 0; cnt < moves.size(); cnt++) {
      Move m = (Move)moves.elementAt(cnt);

      board2 = duplicateBoard(board);

      doMove(board2, m);
      value = boardValue(board2, 0, max);

      if (value >= max) {
        if ((result == null) || (value > max))
          result = m;
        else if (random.nextBoolean())
          result = m;
        max = value;
        if (max == 1000) break;
      }
    }

    return result;
  }

  private void addJumps(int x, int y, char board[][], Move move, Vector result) {
    char c = board[x][y];
    char enemy, enemy2;
    char board2[][];
    Move m;
    if ((c == 'w') || (c == 'W')) {
      enemy = 'b';
      enemy2 = 'B';
    }
    else {
      enemy = 'w';
      enemy2 = 'W';
    }
    // white checkers and super-heroes may move up the board
    if (((c == 'w') || (c == 'W') || (c == 'B')) && (y <= 5)) {
      if ((x >= 2) && ((board[x - 1][y + 1] == enemy) || (board[x - 1][y + 1] == enemy2)) && (board[x - 2][y + 2] == '.')) {
        if (move == null)
          m = new Move(x, y, x - 2, y + 2);
        else {
          m = move.duplicate();
          m.addPosition(x - 2, y + 2);
        }
        result.add(m);
	board2 = duplicateBoard(board);
        board2[x][y] = board2[x - 1][y + 1] = '.';
        board2[x - 2][y + 2] = c;
        addJumps(x - 2, y + 2, board2, m, result);
      }
      if ((x <= 5) && ((board[x + 1][y + 1] == enemy) || (board[x + 1][y + 1] == enemy2)) && (board[x + 2][y + 2] == '.')) {
        if (move == null)
          m = new Move(x, y, x + 2, y + 2);
        else {
          m = move.duplicate();
          m.addPosition(x + 2, y + 2);
        }
        result.add(m);
        board2 = duplicateBoard(board);
        board2[x][y] = board2[x + 1][y + 1] = '.';
        board2[x + 2][y + 2] = c;
        addJumps(x + 2, y + 2, board2, m, result);
      }
    }
    // black checkers and super-heroes may move down the board
    if (((c == 'b') || (c == 'W') || (c == 'B')) && (y >= 2)) {
      if ((x >= 2) && ((board[x - 1][y - 1] == enemy) || (board[x - 1][y - 1] == enemy2)) && (board[x - 2][y - 2] == '.')) {
        if (move == null)
          m = new Move(x, y, x - 2, y - 2);
        else {
          m = move.duplicate();
          m.addPosition(x - 2, y - 2);
        }
        result.add(m);
        board2 = duplicateBoard(board);
        board2[x][y] = board2[x - 1][y - 1] = '.';
        board2[x - 2][y - 2] = c;
        addJumps(x - 2, y - 2, board2, m, result);
      }
      if ((x <= 5) && ((board[x + 1][y - 1] == enemy) || (board[x + 1][y - 1] == enemy2)) && (board[x + 2][y - 2] == '.')) {
        if (move == null)
          m = new Move(x, y, x + 2, y - 2);
        else {
          m = move.duplicate();
          m.addPosition(x + 2, y - 2);
        }
        result.add(m);
        board2 = duplicateBoard(board);
        board2[x][y] = board2[x + 1][y - 1] = '.';
        board2[x + 2][y - 2] = c;
        addJumps(x + 2, y - 2, board2, m, result);
      }
    }
  }

  private Vector getPossibleMoves(char board[][]) {
    int x, y;
    char color2, c;
    Vector result = new Vector();
    if (color == 'w')
      color2 = 'W';
    else
      color2 = 'B';
    for (x = 0; x < 8; x++)
      for (y = 0; y < 8; y++) {
        c = board[x][y];
        if ((c == color) || (c == color2)) {
          // first, check for ordinary (non-killing) moves
          if (((c == 'w') || (c == 'W') || (c == 'B')) && (y < 7)) {
            if ((x > 0) && (board[x - 1][y + 1] == '.'))
              result.add(new Move(x, y, x - 1, y + 1));
            if ((x < 7) && (board[x + 1][y + 1] == '.'))
              result.add(new Move(x, y, x + 1, y + 1));
          }
          if (((c == 'b') || (c == 'B') || (c == 'W')) && (y > 0)) {
            if ((x > 0) && (board[x - 1][y - 1] == '.'))
              result.add(new Move(x, y, x - 1, y - 1));
            if ((x < 7) && (board[x + 1][y - 1] == '.'))
              result.add(new Move(x, y, x + 1, y - 1));
          }
          // then, check for jumps
          addJumps(x, y, board, null, result);
        }
      }
    return result;
  }

}


class Move {
  public int positions[];
  public int length;
  public boolean kill;

  public Move() {
    positions = new int[8];
    length = 0;
  }

  public Move(int fromX, int fromY, int toX, int toY) {
    positions = new int[8];
    length = 0;
    positions[length++] = encodeInt(fromX, fromY);
    positions[length++] = encodeInt(toX, toY);
    if ((toX > fromX + 1) || (toX < fromX - 1))
      kill = true;
    else
      kill = false;
  }

  public void addPosition(int toX, int toY) {
    positions[length++] = encodeInt(toX, toY);
  }

  private char intToRow(int i) {
    if ((i >= 0) && (i < 8))
      return (char)('1' + (char)i);
    else
      return '1';
  }

  private char intToCol(int i) {
    if ((i >= 0) && (i < 8))
      return (char)('a' + (char)i);
    else
      return 'a';
  }

  private int encodeInt(int x, int y) {
    return x + y * 8;
  }

  public int getX(int i) {
    return positions[i] % 8;
  }

  public int getY(int i) {
    return positions[i] / 8;
  }

  public Move duplicate() {
    Move result = new Move();
    result.kill = kill;
    for (int i = 0; i < length; i++)
      result.positions[i] = positions[i];
    result.length = length;
    return result;
  }

  public String toString() {
    String result = "";
    for (int i = 0; i < length; i++)
      result += "" + intToCol(getX(i)) + "" + intToRow(getY(i));
    return result;
  }

}
